## Imbens and Wager: fit quadratic on either side of threshold and multiply by 2
ROT_IW <- function(y, x) {
    4*max(abs(c(lm(y~x+I(x^2), subset=(x>=0))$coefficients[3],
                lm(y~x+I(x^2), subset=(x<0))$coefficients[3])))
}

## Invert AR confidence interval
FRD_ARCI <- function(dl, M) {
    ## Pre-estimate
    f <- RDHonest:::NPRHonest.fit(dl, M=M, kern="uniform", opt.criterion="MSE",
                                  se.method="supplied.var")$coefficients

    p_val <- function(beta) {
        ## RDData object for this beta
        d0 <- dl
        d0$Y <- d0$Y[, 1]-d0$Y[, 2]*beta
        d0$sigma2 <- drop(d0$sigma2 %*% c(1, -beta, -beta, beta^2))
        d0$class <- "SRD"
        r0 <- RDHonest:::NPRHonest.fit(d0, M=M[1]+abs(beta)*M[2],
                                       kern="uniform", opt.criterion="MSE",
                                       se.method="supplied.var")$coefficients
        1-unname(stats::pchisq((r0$estimate/r0$std.error)^2, df = 1,
                               ncp = (r0$maximum.bias/r0$std.error)^2))
    }
    rl <- ru <- f$conf.low
    while(p_val(rl)>0.05) rl <- rl - f$std.error
    while(p_val(ru)<0.05) ru <- f$estimate-(f$estimate-ru)/2
    f$conf.low <- uniroot(function(x) p_val(x)-0.05, c(rl, ru))$root

    rl <- ru <- f$conf.high
    while (p_val(ru)>0.05) ru <- ru + f$std.error
    while (p_val(rl)<0.05) rl <- f$estimate+(rl-f$estimate)/2
    f$conf.high <- uniroot(function(x) p_val(x)-0.05, c(rl, ru))$root

    ## Return full row
    list(coefficients=f)
}

srd_table <- function(x, y, h1, h2, M1, M2, h_rbc=NULL) {
    f1 <- RDHonest::RDHonest(y~x, kern = "uniform", M = 0, h=h1,
                             se.method="EHW")
    f2 <- RDHonest::RDHonest(y~x, kern = "uniform", M = 0, h=h2,
                             se.method="EHW")
    if (is.null(h_rbc)) {
        f3 <- rdrobust::rdrobust(y=y, x=x, kernel="uniform")
    } else {
        f3 <- rdrobust::rdrobust(y=y, x=x, kernel="uniform", h=h_rbc)
    }
    f4 <- RDHonest::RDHonest(y~x, kern = "uniform", M = M1,
                             opt.criterion = "MSE")
    f5 <- RDHonest::RDHonest(y~x, kern = "uniform", M = M2,
                             opt.criterion = "MSE")
    ## Make table
    makerow <- function(f) {
        unlist(f$coefficients[c("estimate", "std.error", "conf.low",
                                "conf.high", "bandwidth", "eff.obs")])
    }

    r <- cbind(makerow(f1), makerow(f2),
               c(f3$coef[3], f3$se[3], f3$ci[3, ], f3$bws[1, 1], sum(f3$N_h)),
               makerow(f4), makerow(f5))
    ## Rescale x to [-1, 1]
    rbind(r, c(rep(NA, 3), c(M1, M2)*max(x^2)))
}

frd_table <- function(x, y, z, h1, h2, M1, M2, h_rbc=NULL) {
    f1 <- RDHonest::RDHonest(y~x | z, kern = "uniform", M = c(0, 0),
                             h=h1, se.method="EHW")
    f2 <- RDHonest::RDHonest(y~x | z, kern = "uniform", M = c(0, 0),
                             h=h2, se.method="EHW")
    if (is.null(h_rbc)) {
        f3 <- rdrobust::rdrobust(y=y, x=z, kernel="uniform", fuzzy=x)
    } else {
        f3 <- rdrobust::rdrobust(y=y, x=z, kernel="uniform", fuzzy=x, h=h_rbc)
    }

    ## Precompute variances
    dl <- RDHonest:::NPRData(data.frame(y=y, x=x, z=z), cutoff=0, class="FRD")
    dl$sigma2 <- matrix(NA, nrow=length(dl$X), ncol=4)
    dl$sigma2[dl$p, ] <- RDHonest:::sigmaNN(dl$X[dl$p], dl$Y[dl$p, ], J=3)
    dl$sigma2[dl$m, ] <- RDHonest:::sigmaNN(dl$X[dl$m], dl$Y[dl$m, ], J=3)

    f4 <- FRD_ARCI(dl, M1)
    f5 <- FRD_ARCI(dl, M2)

    ## Make table
    makerow <- function(f) {
        unlist(f$coefficients[c("estimate", "std.error", "conf.low",
                                "conf.high", "bandwidth", "eff.obs")])
    }
    r <- cbind(makerow(f1), makerow(f2),
               c(f3$coef[3], f3$se[3], f3$ci[3, ], f3$bws[1, 1], sum(f3$N_h)),
               makerow(f4), makerow(f5))
    ## Rescale x to [-1, 1]
    rbind(r, cbind(matrix(NA, nrow=2, ncol=3), cbind(M1, M2)*max(z^2)))
}

write_rd_table <- function(tbl, file, Mname="$M_t$") {
    ## Keep trailing zeros
    nr <- nrow(tbl) # fuzzy vs sharp
    tbl <- rbind(formatC(tbl[1:4, ], format="f", digits=3),
                 formatC(tbl[5:6, ], format="f", digits=0, big.mark=","),
                 ifelse(is.na(tbl[7:nr, ]), "{}",
                        formatC(tbl[7:nr, ], format="f", digits=3)))
    tbl <- rbind(tbl[1:2, ],
                 paste0("(", paste(tbl[3, ], tbl[4, ], sep=","), ")"),
                 tbl[5:nr, ])
    tbl <- tidyr::unite(data=as.data.frame(tbl), col="z", sep = "$ & $")
    nam <- c("Estimate", "SE", "95\\% CI", "Bandwidth", "Eff. obs.",
             if(nr==7) paste("Rescaled", Mname) else c("Rescaled $M_y$",
                                                       "Rescaled $M_t$"))
    write(paste0(nam, " & $", tbl$z, "$\\\\"), file=file)
}


## Adapted from RDHonest::plot_RDscatter
binned_avg <- function(d, avg=10) {
    ## Assume data sorted
    if (avg==Inf) {
        bd <- data.frame(x=unique(d[[2]]),
                         y=unname(coef(lm(d[[1]] ~ 0 + as.factor(d[[2]])))),
                         count=as.vector(table(d[[2]])))
    } else {
        np <- sum(d[[2]] >= 0)
        nm <- sum(d[[2]] < 0)
        ## don't recycle
        maxp <- (np %/% avg) * avg
        rem <- nm %% avg
        bd <- data.frame(
            x=c(colMeans(matrix(d[[2]][d[[2]] < 0][-seq_len(rem)], nrow=avg)),
                colMeans(matrix(d[[2]][d[[2]] >= 0][1:maxp], nrow=avg))),
            y=c(colMeans(matrix(d[[1]][d[[2]] < 0][-seq_len(rem)], nrow=avg)),
                colMeans(matrix(d[[1]][d[[2]] >= 0][1:maxp], nrow=avg))))
        ## if there is a remainder, add it
        if (rem>0)
            bd <- rbind(bd, data.frame(x=mean(d[[2]][seq_len(rem)]),
                                       y=mean(d[[1]][seq_len(rem)])))
        if (maxp+1<=np)
            bd <- rbind(bd, data.frame(x=mean(d[[2]][d[[2]]>=0][(maxp+1):np]),
                                       y=mean(d[[1]][d[[2]]>=0][(maxp+1):np])))
        bd$count <- avg
    }
    bd
}


## Adapted from Noack and Rothe
fitted.points <- function(y, x, M, num_knots=51, eps=0.1) {
    knots <- quantile(x, probs = seq(0, 1, length.out=(num_knots+1)))
    knots <- unname(knots)[-(num_knots+1)]

    C <- cbind(1, x, sapply(knots, FUN=function(c) pmax(0, (x-c))^2))
    A <- cbind(0, 0, sapply(knots, FUN=function(c) 2*as.numeric(x>=c)))
    A <- rbind(A, -A)
    b <- rep(M, 2*length(x))

    Aeq <- matrix(c(0, 0, 2*as.numeric(eps>=knots)), ncol=length(knots)+2)

    c2a <- limSolve::lsei(A=C, B=y, E=Aeq, F=M, G=A, H=-b)
    c2b <- limSolve::lsei(A=C, B=y, E=Aeq, F=-M, G=A, H=-b)
    if (c2a$solutionNorm < c2b$solutionNorm) {
        c2b$X <- c2a$X
    }
    xs <- seq(min(x), max(x), length.out=100)
    X <- cbind(1, xs, sapply(knots, FUN=function(c) pmax(0, (xs-c))^2))
    data.frame(x=xs, y=drop(X %*% c2b$X))
}

visualization <- function(y, x, M, num_knots, eps) {
    rbind(fitted.points(y=y[x<0], x=x[x<0], M, num_knots, -eps),
          fitted.points(y=y[x>=0], x=x[x>=0], M, num_knots, eps))
}
